# -*- coding: utf-8 -*-
# Author Karol Sikora <karol.sikora@laboratorium.ee>, (c) 2012
# Author Michal Nowotka <mmmnow@gmail.com>, (c) 2013-2014

try:
    import cPickle as pickle
except ImportError:
    import pickle
import base64
import pymongo
from datetime import datetime, timedelta
from django.core.cache.backends.base import BaseCache
import zlib
import logging

#-----------------------------------------------------------------------------------------------------------------------

MAX_SIZE = 16000000

#-----------------------------------------------------------------------------------------------------------------------

class MongoDBCache(BaseCache):

    def __init__(self, location, params):
        BaseCache.__init__(self, params)
        self.location = location
        options = params.get('OPTIONS', {})
        self._host = options.get('HOST', 'localhost')
        self._port = options.get('PORT', 27017)
        self._database = options.get('DATABASE', 'django_cache')
        self._rshosts = options.get('RSHOSTS')
        self._rsname = options.get('RSNAME')
        self._user = options.get('USER', None)
        self._password = options.get('PASSWORD', None)
        self._socket_timeout_ms = options.get('SOCKET_TIMEOUT_MS', None)
        self._connect_timeout_ms = options.get('CONNECT_TIMEOUT_MS', 20000)
        self.compression_level = options.get('COMPRESSION_LEVEL', 0)
        self._tag_sets = options.get('TAG_SETS', None)
        self._read_preference = options.get("READ_PREFERENCE")
        self._collection = location
        self.log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------------------------------------------------

    def add(self, key, value, timeout=None, version=None):
        key = self.make_key(key, version)
        self.validate_key(key)
        self._base_set('add', key, value, timeout)

#-----------------------------------------------------------------------------------------------------------------------

    def set(self, key, value, timeout=None, version=None):
        key = self.make_key(key, version)
        self.validate_key(key)
        self._base_set('set', key, value, timeout)

#-----------------------------------------------------------------------------------------------------------------------

    def _base_set(self, mode, key, value, timeout=None):
        if not timeout:
            timeout = self.default_timeout
        now = datetime.utcnow()
        expires = now + timedelta(seconds=timeout)
        coll = self._get_collection()
        encoded = self._encode(value)
        document_size = len(encoded)
        count = coll.count()
        if count > self._max_entries:
            self._cull()
        data = coll.find_one({'_id': key})
        if data and (mode == 'set' or
                (mode == 'add' and data['expires'] > now)):
            coll.update({'_id': data['_id']}, {'$set': {'expires': expires}}, safe=True)
        else:
            if document_size <= MAX_SIZE:
                coll.insert({'_id': key, 'data': encoded, 'expires': expires}, safe=True)
            else:
                chunks = []
                for i in xrange(0, document_size, MAX_SIZE):
                    chunk = encoded[i:i+MAX_SIZE]
                    aux_key = self.make_key(chunk)
                    coll.insert({'_id': aux_key, 'data': chunk}, safe=True)
                    chunks.append(aux_key)
                coll.insert({'_id': key, 'chunks': chunks, 'expires': expires}, safe=True)

#-----------------------------------------------------------------------------------------------------------------------

    def _decode(self, data):
        return pickle.loads(zlib.decompress(base64.decodestring(data)))

#-----------------------------------------------------------------------------------------------------------------------

    def _encode(self, data):
        return base64.encodestring(zlib.compress(pickle.dumps(data, pickle.HIGHEST_PROTOCOL), self.compression_level))

#-----------------------------------------------------------------------------------------------------------------------

    def get(self, key, default=None, version=None):
        coll = self._get_collection()
        key = self.make_key(key, version)
        self.validate_key(key)
        now = datetime.utcnow()
        data = coll.find_one({'_id': key})
        if not data:
            return default
        if data['expires'] < now:
            coll.remove(data['_id'])
            return default
        if data.get('data', None):
            raw = data['data']
        elif data.get('chunks', None):
            raw = ''
            chunks = data.get('chunks')
            for chunk in chunks:
                raw += coll.find_one({'_id': chunk})['data']
        else:
            return default
        return self._decode(raw)

#-----------------------------------------------------------------------------------------------------------------------

    def get_many(self, keys, version=None):
        coll = self._get_collection()
        now = datetime.utcnow()
        out = {}
        parsed_keys = {}
        to_remove = []
        for key in keys:
            pkey = self.make_key(key, version)
            self.validate_key(pkey)
            parsed_keys[pkey] = key
        data = coll.find({'_id': {'$in': parsed_keys.keys()}})
        for result in data:
            if result['expires'] < now:
                to_remove.append(result['_id'])
            elif result.get('data', None):
                raw = result['data']
            elif result.get('chunks', None):
                raw = ''
                chunks = result.get('chunks')
                for chunk in chunks:
                    raw += coll.find_one({'_id': chunk})['data']
            out[parsed_keys[result['_id']]] = self._decode(raw)
        if to_remove:
            coll.remove({'_id': {'$in': to_remove}})
        return out

#-----------------------------------------------------------------------------------------------------------------------

    def delete(self, key, version=None):
        key = self.make_key(key, version)
        self.validate_key(key)
        coll = self._get_collection()
        coll.remove({'_id': key})

#-----------------------------------------------------------------------------------------------------------------------

    def has_key(self, key, version=None):
        coll = self._get_collection()
        key = self.make_key(key, version)
        self.validate_key(key)
        data = coll.find_one({'_id': key, 'expires': {'$gt': datetime.utcnow()}})
        return data is not None

#-----------------------------------------------------------------------------------------------------------------------

    def clear(self):
        coll = self._get_collection()
        coll.remove(None)

#-----------------------------------------------------------------------------------------------------------------------

    def _cull(self):
        if self._cull_frequency == 0:
            self.clear()
            return
        coll = self._get_collection()
        coll.remove({'expires': {'$lte': datetime.utcnow()}})
        #TODO: implement more agressive cull

#-----------------------------------------------------------------------------------------------------------------------

    def _get_collection(self):
        if not getattr(self, '_coll', None):
            self._initialize_collection()
        return self._coll

#-----------------------------------------------------------------------------------------------------------------------

    def _initialize_collection(self):
        try:
            from gevent import monkey
            monkey.patch_socket()
        except ImportError:
            pass

        if self._rsname:
            self.connection = pymongo.MongoReplicaSetClient(self._rshosts, replicaSet=self._rsname,
                read_preference=self._read_preference, socketTimeoutMS=self._socket_timeout_ms,
                connectTimeoutMS=self._connect_timeout_ms, tag_sets=self._tag_sets)
        else:
            self.connection = pymongo.Connection(self._host, self._port)

        self._db = self.connection[self._database]
        if self._user and self._password:
            self._db.authenticate(self._user, self._password)
        self._coll= self._db[self._collection]

#-----------------------------------------------------------------------------------------------------------------------


__author__ = 'mnowotka'

from django.db import models
from django.db.models import Field
from django.db.models import NOT_PROVIDED
import base64
try:
    import cx_Oracle as oracle
except ImportError:
    oracle = None
try:
    import psycopg2
except ImportError:
    psycopg2 = None
from django.utils.translation import ugettext_lazy as _
from django.core import exceptions
from django.utils.datastructures import DictWrapper
from django.utils import six

#-----------------------------------------------------------------------------------------------------------------------

def _adjust_keywords(kwds):
    required = kwds.get('required', False)
    kwds['null'] = not required
    kwds['blank'] = not required
    if 'required' in kwds:
        del kwds['required']
    if 'choices' in kwds:
        kwds['choices'] = [(a, a) for a in kwds['choices']]
    return kwds

#-----------------------------------------------------------------------------------------------------------------------

class Blob(str):
    pass

#-----------------------------------------------------------------------------------------------------------------------

class BlobField(six.with_metaclass(models.SubfieldBase, models.TextField)):
    description = "Stores raw binary data"

    def __init__(self, *args, **kwds):
        kwds = _adjust_keywords(kwds)
        super(BlobField, self).__init__(*args, **kwds)

    def db_type(self, connection):
        if connection.vendor == 'mysql':
            return 'BLOB'
        if connection.vendor == 'oracle':
            return 'BLOB'
        if connection.vendor == 'postgresql':
            return 'bytea'
        if connection.vendor == 'sqlite':
            return 'BLOB'

    def get_internal_type(self):
        return "BlobField"

    def get_db_prep_value(self, value, connection=None, prepared=False):
        return value

    def get_db_prep_save(self, value, connection):
        if connection:
            if connection.vendor == 'oracle':
                curs = connection.cursor()
                blob = curs.var(oracle.BLOB)
                blob.setvalue(0, value)
                return blob
            if connection.vendor == 'postgresql':
                return psycopg2.Binary(value)
        return value

    def value_to_string(self, obj):
        return base64.b64encode(self._get_val_from_obj(obj))

    def to_python(self, value):
        if isinstance(value, Blob):
            return value
        if isinstance(value, basestring):
            try:
                unicode(value)
                return Blob(base64.b64decode(value))
            except UnicodeDecodeError:
                return Blob(value)
        return Blob(value)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblTextField(models.TextField):
    empty_strings_allowed = False

    def db_type(self, connection):
        if connection.vendor == 'oracle':
            return 'CLOB'
        return super(ChemblTextField, self).db_type(connection)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblCharField(models.CharField):
    empty_strings_allowed = False

    def __init__(self, *args, **kwargs):
        self.novalidate = kwargs.pop('novalidate', False)
        self.novalidate_default = kwargs.pop('novalidate_default', False)
        super(ChemblCharField, self).__init__(*args, **kwargs)


    def db_type(self, connection):

        data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
        default = ''
        if self.default != NOT_PROVIDED and not self.novalidate_default:
            default = (" DEFAULT '%s' " % str(self.default))

        if connection.vendor == 'postgresql':
            if self.max_length >= 2712:
                self.db_index = False
                self._unique = False
            type = 'varchar(%(max_length)s)'
            type += default
            if self.choices and not self.novalidate:
                choices = ', '.join(map(lambda x: "'%s'" % x[0].replace("'", "''"), self.choices))
                type += ' CHECK (%(column)s IN (' + choices + ')) '
            return (type % data)

        if connection.vendor == 'mysql':
            if self.max_length > 767:
                self.db_index = False
                self._unique = False
            type = 'varchar(%(max_length)s)'
            type += default
            return (type % data)

        if connection.vendor == 'sqlite':
            type = 'varchar(%(max_length)s)'
            type += default
            return (type % data)

        if connection.vendor == 'oracle':
            type = 'VARCHAR2(%(max_length)s BYTE)'
            type += default
            if self.choices and not self.novalidate:
                choices = ', '.join(map(lambda x: "'%s'" % x[0].replace("'", "''"), self.choices))
                type += ' CHECK (%(qn_column)s IN (' + choices + ')) '
            return (type % data)
        return super(ChemblCharField, self).db_type(connection)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblDateField(models.DateField):

    def db_type(self, connection):
        type = 'DATE'
        if connection.vendor == 'oracle':
            if self.default != NOT_PROVIDED:
                if hasattr(self.default, '__call__') and self.default.__name__ == 'today':
                    type += ' DEFAULT sysdate '
            return type
        if connection.vendor == 'postgresql':
            if self.default != NOT_PROVIDED:
                if hasattr(self.default, '__call__') and self.default.__name__ == 'today':
                    type += ' DEFAULT CURRENT_DATE '
            return type
        if connection.vendor == 'sqlite':
            if self.default != NOT_PROVIDED:
                if hasattr(self.default, '__call__') and self.default.__name__ == 'today':
                    type += " (datetime('now','localtime')) "
            return type
        return super(ChemblDateField, self).db_type(connection)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblNoLimitDecimalField(models.DecimalField):

    def __init__(self, verbose_name=None, name=None,  **kwargs):
        super(ChemblNoLimitDecimalField, self).__init__(verbose_name=verbose_name, name=name, max_digits=38,
            decimal_places=19, **kwargs)

    def db_type(self, connection):
        if connection.vendor == 'oracle':
            return 'NUMBER'
        if connection.vendor == 'postgresql':
            return 'NUMERIC'
        if connection.vendor == 'mysql':
            return 'NUMERIC(64,30)'
        return super(ChemblNoLimitDecimalField, self).db_type(connection)

    def format_number(self, value):
        return str(value)

    def get_db_prep_save(self, value, connection):
        if value is not None:
            return self.format_number(value)
        return value

#-----------------------------------------------------------------------------------------------------------------------

class ChemblPositiveDecimalField(models.DecimalField):

    def db_type(self, connection):
        if connection.vendor == 'oracle':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'NUMBER(%(max_digits)s, %(decimal_places)s) CHECK (%(qn_column)s >= 0)'
            return (type % data)

        if connection.vendor == 'postgresql':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type =  'numeric(%(max_digits)s, %(decimal_places)s) CHECK ("%(column)s" >= 0)'
            return (type % data)

        return super(ChemblPositiveDecimalField, self).db_type(connection)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblIntegerField(models.IntegerField):

    def get_internal_type(self):
        return "ChemblIntegerField"

    def __init__(self, length, *args, **kwargs):
        self.length = length
        super(ChemblIntegerField, self).__init__(*args, **kwargs)

    def db_type(self, connection):

        default = ''
        if self.default != NOT_PROVIDED:
            default = (' DEFAULT %s ' % str(int(self.default)))

        if connection.vendor == 'oracle':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'NUMBER(%s,0)' % (self.length)
            type += default
            if self.choices:
                choices = ', '.join(map(lambda x: str(x[0]), self.choices))
                type += ' CHECK (%(qn_column)s IN (' + choices + ')) '
            return (type % data)

        if connection.vendor == 'postgresql':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'integer'
            if self.length <= 4:
                type = 'smallint'
            if self.length > 9:
                type =  'bigint'
            type += default
            if self.choices:
                choices = ', '.join(map(lambda x: str(x[0]), self.choices))
                type += ' CHECK (%(column)s IN (' + choices + ')) '
            return (type % data)

        if connection.vendor == 'mysql':
            type = 'int(%s)' % (self.length)
            return type + default

        if connection.vendor == 'sqlite':
            type = 'integer'
            if self.length <= 4:
                type = 'smallint'
            if self.length > 9:
                type =  'bigint'
            return type + default

        return super(ChemblIntegerField, self).db_type(connection)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblNullBooleanField(models.IntegerField):

    def get_internal_type(self):
        return "ChemblNullBooleanField"

    def db_type(self, connection):
        default = ''
        if self.default != NOT_PROVIDED:
            default = (' DEFAULT %s ' % str(int(self.default)))
        if connection.vendor == 'oracle':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'NUMBER(1,0)'
            type += default
            type += ' CHECK (%(qn_column)s IN (0,1,-1)) '
            return (type % data)
        if connection.vendor == 'postgresql':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'smallint'
            type += default
            type += ' CHECK ("%(column)s" in (0,1,-1))'
            return (type % data)
        if connection.vendor == 'mysql':
            type = 'int(1)'
            type += default
            return type
        return super(ChemblNullBooleanField, self).db_type(connection)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblNullableBooleanField(models.NullBooleanField):

    def __init__(self, *args, **kwargs):
        kwargs['null'] = True
        kwargs['blank'] = True
        Field.__init__(self, *args, **kwargs)

    def get_db_prep_value(self, value, connection=None, prepared=False):
        if connection.vendor == 'postgresql':
            if value is None:
                return None
            if value:
                return 1
            return 0
        return super(ChemblNullableBooleanField, self).get_db_prep_value(value, connection, prepared)

    def db_type(self, connection):
        if connection.vendor == 'oracle':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'NUMBER(1)'
            if self.default != NOT_PROVIDED:
                type += (' DEFAULT %s ' % str(int(self.default)))
            type += ' CHECK ((%(qn_column)s IN (0,1)) OR (%(qn_column)s IS NULL)) '
            return (type % data)
        if connection.vendor == 'postgresql':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'smallint'
            if self.default != NOT_PROVIDED:
                type += (' DEFAULT %s ' % str(int(self.default)))
            type += ' CHECK ((%(column)s IN (0,1)) OR (%(column)s IS NULL)) '
            return (type % data)
        if connection.vendor == 'mysql':
            type = 'bool'
            if self.default != NOT_PROVIDED:
                type += (' DEFAULT %s ' % str(int(self.default)))
            return type
        if connection.vendor == 'sqlite':
            type = 'BOOL'
            if self.default != NOT_PROVIDED:
                type += (' DEFAULT %s ' % str(int(self.default)))
            return type
        return super(ChemblNullableBooleanField, self).db_type(connection)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblBooleanField(models.BooleanField):

    def __init__(self, *args, **kwargs):
        Field.__init__(self, *args, **kwargs)

    def get_db_prep_value(self, value, connection=None, prepared=False):
        if connection.vendor == 'postgresql':
            if value:
                return 1
            return 0
        return super(ChemblBooleanField, self).get_db_prep_value(value, connection, prepared)

    def db_type(self, connection):
        if connection.vendor == 'oracle':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'NUMBER(1)'
            if self.default != NOT_PROVIDED:
                type += (' DEFAULT %s ' % str(int(self.default)))
            type += ' CHECK (%(qn_column)s IN (0,1)) '
            return (type % data)
        if connection.vendor == 'postgresql':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'smallint'
            if self.default != NOT_PROVIDED:
                type += (' DEFAULT %s ' % str(int(self.default)))
            type += ' CHECK (%(column)s IN (0,1)) '
            return (type % data)
        if connection.vendor == 'mysql':
            type = 'bool'
            if self.default != NOT_PROVIDED:
                type += (' DEFAULT %s ' % str(int(self.default)))
            return type
        if connection.vendor == 'sqlite':
            type = 'BOOL'
            if self.default != NOT_PROVIDED:
                type += (' DEFAULT %s ' % str(int(self.default)))
            return type
        return super(ChemblBooleanField, self).db_type(connection)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblPositiveIntegerField(models.IntegerField):

    def get_internal_type(self):
        return "PositiveIntegerField"

    def formfield(self, **kwargs):
        defaults = {'min_value': 0}
        defaults.update(kwargs)
        return super(ChemblPositiveIntegerField, self).formfield(**defaults)

    def __init__(self, length, *args, **kwargs):
        self.length = length
        super(ChemblPositiveIntegerField, self).__init__(*args, **kwargs)

    def db_type(self, connection):

        default = ''
        choices = ''

        if self.default != NOT_PROVIDED:
            default = (' DEFAULT %s ' % str(int(self.default)))

        if connection.vendor == 'oracle':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'NUMBER(%s,0)' % (self.length)
            type += default
            if self.choices:
                choices = ' AND %(qn_column)s IN ('+ ', '.join(map(lambda x: str(x[0]), self.choices)) + ' ) '
            type += ' CHECK (%(qn_column)s >= 0' + choices + ') '
            return (type % data)

        if connection.vendor == 'postgresql':
            data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
            type = 'integer'
            if self.length <= 4:
                type =  'smallint'
            if self.length > 9:
                type = 'bigint'
            type += default
            if self.choices:
                choices = ' AND %(column)s IN ('+ ', '.join(map(lambda x: str(x[0]), self.choices)) + ' ) '
            type += ' CHECK (%(column)s >= 0' + choices + ') '
            return (type % data)

        if connection.vendor == 'mysql':
            type =  'int(%s) UNSIGNED' % (self.length)
            type += default
            return type

        if connection.vendor == 'sqlite':
            type = 'integer'
            if self.length <= 4:
                type = 'smallint'
            if self.length > 9:
                type =  'bigint'
            type += ' unsigned '
            type += default
            return type

        return super(ChemblPositiveIntegerField, self).db_type(connection)

#-----------------------------------------------------------------------------------------------------------------------

class ChemblAutoField(Field):
    description = _("Integer")

    empty_strings_allowed = False
    default_error_messages = {
        'invalid': _(u"'%s' value must be an integer."),
        }

    def __init__(self, length, *args, **kwargs):
        assert kwargs.get('primary_key', False) is True,\
        "%ss must have primary_key=True." % self.__class__.__name__
        kwargs['blank'] = True
        self.length = length
        Field.__init__(self, *args, **kwargs)

    def get_internal_type(self):
        return "AutoField"

    def db_type(self, connection):
        if connection.vendor == 'oracle':
            return 'NUMBER(%s,0)' % (self.length)
        if connection.vendor == 'postgresql':
            if self.length <= 4:
                return 'smallint'
            if self.length <= 9:
                return 'integer'
            return 'bigint'
        if connection.vendor == 'mysql':
            return 'int(%s)' % (self.length)
        if connection.vendor == 'sqlite':
            type = 'integer'
            if self.length <= 4:
                type = 'smallint'
            if self.length > 9:
                type =  'bigint'
            return type
        return super(ChemblAutoField, self).db_type(connection)

    def to_python(self, value):
        if value is None:
            return value
        try:
            return int(value)
        except (TypeError, ValueError):
            msg = self.error_messages['invalid'] % str(value)
            raise exceptions.ValidationError(msg)

    def validate(self, value, model_instance):
        pass

    def get_prep_value(self, value):
        if value is None:
            return None
        return int(value)

    def contribute_to_class(self, cls, name):
        assert not cls._meta.has_auto_field,\
        "A model can't have more than one AutoField."
        super(ChemblAutoField, self).contribute_to_class(cls, name)
        cls._meta.has_auto_field = True
        cls._meta.auto_field = self

    def formfield(self, **kwargs):
        return None

#-----------------------------------------------------------------------------------------------------------------------


__author__ = 'mnowotka'

from django.db import models
from django.db import connections
from django.db.models.query import QuerySet

#-----------------------------------------------------------------------------------------------------------------------

class CompoundMolsMixin(object):

    def get_column(self, name):
        return filter(lambda x: x.name == name, self.model._meta.fields)[0].db_column or name

    def similar_to(self, structure, similarity_index):
        ctab_column = self.get_column('ctab')
        molregno_column = self.get_column('molecule')
        try:
            sim = int(similarity_index)
        except ValueError:
            raise ValueError('similarity_index must be integer from range (50,100)')
        if sim < 50 or sim > 100:
            raise ValueError('similarity_index must be integer from range (50,100)')

        connection = connections[self._db or 'default']
        if connection.vendor == 'oracle':
            return self.extra(select={'similarity': "TO_NUMBER (molsim (" + ctab_column + ",%s, 'normal'))"},
                select_params=(structure,),
                where=["molsim (" + ctab_column + ", %s, 'normal') BETWEEN %s AND '100'"],
                params=('smiles:' + structure, similarity_index))
        if connection.vendor == 'postgresql':
            cursor = connection.cursor()
            cursor.execute("select 'c1ccccc1O'::mol;") # dirty hack but what can I do...
            cursor.execute('set rdkit.tanimoto_threshold=%s;', (sim / 100.0,))
            ret = self.extra(select={'similarity': "tanimoto_sml(torsionbv_fp(%s), torsionbv)"},
                select_params=(structure,),
                where=["fps_rdkit.molregno = " + self.model._meta.db_table + "." + molregno_column +
                       " and torsionbv_fp(%s) %% torsionbv and tanimoto_sml(torsionbv_fp(%s), torsionbv) between %s and 1.0"],
                tables=['fps_rdkit'],
                order_by=['-similarity'],
                params=[structure, structure, (sim / 100.0)])
            cursor.execute('set rdkit.tanimoto_threshold=0.5;')
            return ret
        else:
            raise NotImplementedError

    def with_substructure(self, structure):
        ctab_column = self.get_column('ctab')
        connection = connections[self._db or 'default']
        if connection.vendor == 'oracle':
            return self.extra(where=["(sss(" + ctab_column + ",%s,'ignore=all')=1)"], params=('smiles:' + structure,))
        if connection.vendor == 'postgresql':
            return self.extra(where=[ctab_column + "@>%s"], params=(structure,))
        else:
            raise NotImplementedError

    def flexmatch(self, structure):
        ctab_column = self.get_column('ctab')
        connection = connections[self._db or 'default']
        if connection.vendor == 'oracle':
            return self.extra(where=["(flexmatch(" + ctab_column + ",%s,'ignore=all')=1)"], params=('smiles:' + structure,))
        if connection.vendor == 'postgresql':
            return self.extra(where=[ctab_column + "@=%s"], params=(structure,))
        else:
            raise NotImplementedError

#-----------------------------------------------------------------------------------------------------------------------

class CompoundMolsQuerySet(QuerySet,CompoundMolsMixin):
    pass

#-----------------------------------------------------------------------------------------------------------------------

class CompoundMolsManager(models.Manager, CompoundMolsMixin):
    def get_query_set(self):
        return CompoundMolsQuerySet(self.model, using=self._db)

#-----------------------------------------------------------------------------------------------------------------------


from django.db.backends.oracle.introspection import DatabaseIntrospection as OracleDatabaseIntrospection
import cx_Oracle
import re

#-----------------------------------------------------------------------------------------------------------------------

IN_REGEX = re.compile(r'(\S+) in (\([^\)]+\))')
IS_NOT_NULL_REGEX = re.compile(r'(\S+) is not null')
IS_POSITIVE_REGEX = re.compile(r'(\S+) >= 0')
DISTINCT_CUTOFF = 10

#-----------------------------------------------------------------------------------------------------------------------

class DatabaseIntrospection(OracleDatabaseIntrospection):

    data_types_reverse = {
        cx_Oracle.CLOB: 'ChemblTextField',
        cx_Oracle.NCLOB: 'ChemblTextField',
        cx_Oracle.BLOB: 'BlobField',
        cx_Oracle.DATETIME: 'ChemblDateField',
        cx_Oracle.STRING: 'ChemblCharField',
        cx_Oracle.FIXED_CHAR: 'ChemblCharField',

        cx_Oracle.NUMBER: 'DecimalField',
        cx_Oracle.TIMESTAMP: 'DateTimeField',
        }

#-----------------------------------------------------------------------------------------------------------------------

    def get_field_type(self, data_type, description):
        #print description
        constraints = description[7]
        precision = description[4]
        scale = description[5]
        ins = [c for c in constraints if IN_REGEX.match(c)]
        if data_type == cx_Oracle.NUMBER:
            if precision == 0 and scale == (-127):
                return 'ChemblNoLimitDecimalField'
            if precision == 1 and scale == 0:
                if ins and any(map(lambda x : IN_REGEX.search(x).groups()[1] == '(0,1)', ins)):
                        if description[6] == 0 or any(map(lambda x : IS_NOT_NULL_REGEX.match(x), constraints)):
                            return 'ChemblBooleanField'
                        return 'ChemblNullableBooleanField'
                elif ins and any(map(lambda x : IN_REGEX.search(x).groups()[1] == '(-1,0,1)', ins)):
                    return 'ChemblNullBooleanField'
            if description[5] == 0:
                if any(map(lambda x:IS_POSITIVE_REGEX.match(x), constraints)) or any(map(lambda x :
                                                            IN_REGEX.search(x).groups()[1].startswith('(0,'), ins)):
                    return 'ChemblPositiveIntegerField'
                return 'ChemblIntegerField'
            if any(map(lambda x:IS_POSITIVE_REGEX.match(x), constraints)):
                return 'ChemblPositiveDecimalField'
            return 'DecimalField'
        else:
            return super(DatabaseIntrospection, self).get_field_type(
                data_type, description)

#-----------------------------------------------------------------------------------------------------------------------

    def get_table_description(self, cursor, table_name):
        "Returns a description of the table, with the DB-API cursor.description interface."
        cursor.execute("SELECT * FROM %s WHERE ROWNUM < 2" % self.connection.ops.quote_name(table_name))
        description = []
        for desc in cursor.description:
            description.append((desc[0].lower(),) + desc[1:])
        #print str(description)
        return description

#-----------------------------------------------------------------------------------------------------------------------

    def get_sequence_incrementing_triggers(self, cursor, table_name):
        cursor.execute("""
    select trigger_name
    from USER_TRIGGERS
    where table_name=%s
      and status='ENABLED'
      and triggering_event='INSERT';""", [table_name.upper()])

        triggers = []
        for row in cursor.fetchall():
            triggers.append(row[0])
        return triggers

#-----------------------------------------------------------------------------------------------------------------------

    def get_distinct_values(self, cursor, table_name, column_name):
        values = []
        cursor.execute("""
    select distinct %s
    from %s
    order by %s;""" % (column_name.upper(), table_name.upper(), column_name.upper()))
        for row in cursor.fetchall():
            values.append(row[0])
        return values

#-----------------------------------------------------------------------------------------------------------------------

    def is_positive(self, cursor, table_name, column_name):
        values = []
        cursor.execute("""
    select min(%s)
    from %s;""" % (column_name.upper(), table_name.upper()))
        val = cursor.fetchone()
        return val[0] >= 0

#-----------------------------------------------------------------------------------------------------------------------

    def get_unique_together(self, cursor, table_name):
        cursor.execute("""
    SELECT i.index_name,
       c.column_name,
       d.data_length
  FROM user_indexes i,
       user_cons_columns c,
       user_tab_cols d
 WHERE i.table_name  = %s
   AND i.uniqueness  = 'UNIQUE'
   AND i.index_name  = c.constraint_name
   AND i.table_owner = c.owner
   AND i.table_name  = c.table_name
   AND c.column_name = d.column_name
   AND c.table_name = d.table_name
   AND c.constraint_name IN (SELECT constraint_name FROM user_cons_columns WHERE position = 2)
 ORDER BY i.index_name, c.position;""", [table_name.upper()])

        uniques = {}
        sum = 0
        for row in cursor.fetchall():
            sum += int(row[2])
            if row[0] in uniques:
                uniques[row[0]].append(row[1])
            else:
                uniques[row[0]] = [row[1]]
        if sum >= 767:
            print '\033[0m\033[1;%dmWarning [%s]: Uniqe together index for these columns may be to long for MySQL capabilities: %s' % (34, 5, str(uniques))
        return uniques

#-----------------------------------------------------------------------------------------------------------------------

    def get_arity(self, cursor, table_name):
        table_name = table_name.upper()
        cursor.execute("""
SELECT  column_name, num_distinct, nullable, data_type, char_length, column_id
FROM    user_tab_columns t
WHERE   t.table_name = %s
AND data_type NOT IN ('BLOB', 'CLOB', 'NCLOB', 'DATETIME', 'TIMESTAMP', 'DATE')
AND (data_scale = 0 OR data_scale is null);""", [table_name])

        arity = {}
        for row in cursor.fetchall():
            arity[row[0].lower()] = (row[1], row[2], row[3], row[4], row[5] - 1)
        return arity

#-----------------------------------------------------------------------------------------------------------------------

    def get_choices(self, constraints, arity, cursor, table_name, comments):
        ar = filter(lambda x: arity[x][0] > 1 and arity[x][0] <= DISTINCT_CUTOFF, arity)
        ins = [c for c in constraints if IN_REGEX.match(c)]
        res = map(lambda x: (IN_REGEX.search(x).groups()[0], IN_REGEX.search(x).groups()[1][1:-1]), ins)
        ret = {}
        null_flags = []
        flags = []
        tri_state_flags = []
        for r in res:
            choices = []
            try:
                for choice in r[1].split(','):
                    ch = choice.strip()
                    if ch.startswith("'"):
                        ch = str(ch[1:-1].upper())
                    else:
                        ch = int(ch)
                    choices.append((ch, str(ch)))
            except Exception as e:
                print constraints
                print "exception for choices: --->%s<---- (%s)" % (r[1], res)
                raise e
            ret[r[0]] = tuple(choices)

        choices_without_constraints = filter(lambda x: x not in ret.keys() ,ar)
        for ch in choices_without_constraints:
            distinct = self.get_distinct_values(cursor, table_name, ch)
            no_null_distinct = [x for x in distinct if x != None]
            if not no_null_distinct:
                print "\033[0m\033[1;%dmWarning [%s]: column %s.%s is empty\033[0m" % (30, 1, table_name, ch)
                continue
            if 'CHAR' in arity[ch][2]:
                max_len  = max(map(lambda x:len(x) if x else 0, no_null_distinct))
                if  max_len *2 < arity[ch][3]:
                    print "\033[0m\033[1;%dmWarning [%s]: Longest string in %s.%s is more than two times shorter (%s) than maximal allowed length (%s)\033[0m" % (35, 6, table_name, ch, max_len, arity[ch][3])
            if len(distinct) == 2 and len(no_null_distinct) == 1:
                if no_null_distinct[0] == 1:
                    null_flags.append(ch)
                    if not ch.endswith('_flag') and not comments.get(ch, '').lower().startswith('flag'):
                        print "\033[0m\033[1;%dmWarning [%s]: column %s.%s has only two distinct values (NULL, %s) - should it be a flag?\033[0m" % (36, 7, table_name, ch, str(no_null_distinct[0]))
                else:
                    print "\033[0m\033[1;%dmWarning [%s]: column %s.%s has only two distinct values (NULL, %s) - should it be a flag?\033[0m" % (36, 7, table_name, ch, str(no_null_distinct[0]))
                continue
            if len(no_null_distinct) > 2:
                if len(no_null_distinct) == 3 and no_null_distinct[0] == -1 and no_null_distinct[1] == 0 and no_null_distinct[2] == 1:
                    tri_state_flags.append(ch)
                    if ch.endswith('_flag') and not comments.get(ch, '').lower().startswith('flag'):
                        print "\033[0m\033[1;%dmWarning [%s]: column %s.%s has only %s distinct values (%s) but no constraint.\033[0m" % (37, 8, table_name, ch, len(no_null_distinct), ', '.join(map(lambda x: str(x),no_null_distinct)))
                else:
                    if not ch.endswith(('_type', '_lookup', '_version', '_by')):
                        print "\033[0m\033[1;%dmWarning [%s]: column %s.%s has only %s distinct values (%s) but no constraint.\033[0m" % (37, 8, table_name, ch, len(no_null_distinct), ', '.join(map(lambda x: str(x),no_null_distinct)))
                    if 'CHAR' in arity[ch][2]:
                        ret[ch] = tuple(map(lambda x: (str(x), str(x)), no_null_distinct))
                    else:
                        ret[ch] = tuple(map(lambda x: (x, str(x)), no_null_distinct))
            elif len(no_null_distinct) == 2:
                if no_null_distinct[0] == 0 and no_null_distinct[1] == 1:
                    if arity[ch][1] == 'Y':
                        null_flags.append(ch)
                    else:
                        flags.append(ch)
                    if not ch.endswith('_flag') and not comments.get(ch, '').lower().startswith('flag'):
                        print "\033[0m\033[1;%dmWarning [%s]: column %s.%s has only two distinct values (%s) - should it be a flag?\033[0m" % (30, 9, table_name, ch, ', '.join(map(lambda x: str(x),no_null_distinct)))
                else:
                    print "\033[0m\033[1;%dmWarning [%s]: column %s.%s has only two distinct values (%s) - should it be a flag?\033[0m" % (30, 9, table_name, ch, ', '.join(map(lambda x: str(x),no_null_distinct)))
                    if 'CHAR' in arity[ch][2]:
                        ret[ch] = tuple(map(lambda x: (str(x), str(x)), no_null_distinct))
                    else:
                        ret[ch] = tuple(map(lambda x: (x, str(x)), no_null_distinct))
        return ret, null_flags, flags, tri_state_flags

#-----------------------------------------------------------------------------------------------------------------------

    def get_defaults(self, cursor, table_name):
        table_name = table_name.upper()
        cursor.execute("""
    SELECT COLUMN_NAME, DATA_DEFAULT, DATA_TYPE
    FROM USER_TAB_COLUMNS
    WHERE TABLE_NAME = %s AND DATA_DEFAULT IS NOT NULL;""", [table_name])

        defaults = {}
        for row in cursor.fetchall():
            if row[1] == 'NULL':
                continue
            if row[2] == 'NUMBER':
                try:
                    defaults[row[0].lower()] = int(row[1])
                except:
                    continue
            else:
                default = row[1].strip()
                if default.startswith("'") and default.endswith("'"):
                    default = default[1:-1]
                defaults[row[0].lower()] = default
        return defaults

#-----------------------------------------------------------------------------------------------------------------------

    def get_contraints(self, cursor, table_name):
        table_name = table_name.upper()
        cursor.execute("""
    select search_condition
    from user_constraints
    where table_name=%s and constraint_type='C';""", [table_name])

        constraints = []
        for row in cursor.fetchall():
            constraints.append(row[0].lower())
        return constraints

#-----------------------------------------------------------------------------------------------------------------------

    def get_comments(self, cursor, table_name):
        table_name = table_name.upper()
        cursor.execute("""
    SELECT lower(tc.column_name) as column_name
    ,      cc.comments
    FROM   user_col_comments cc
    JOIN   user_tab_columns  tc on  cc.column_name = tc.column_name
                                and cc.table_name  = tc.table_name
    WHERE  cc.table_name = %s AND cc.comments IS NOT NULL;""", [table_name])

        comments = {}
        for row in cursor.fetchall():
            if 'deprecated' in row[1].lower():
                print "\033[0m\033[1;%dmWarning [%s]: column %s.%s may be deprecated \033[0m" % (32, 10, table_name, row[0])
            comments[row[0]] = (row[1])
        return comments

#-----------------------------------------------------------------------------------------------------------------------

    def get_relations(self, cursor, table_name):
        """
        Returns a dictionary of {field_index: (field_index_other_table, other_table)}
        representing all relationships to the given table. Indexes are 0-based.
        """
        table_name = table_name.upper()
        cursor.execute("""
    SELECT ta.column_id - 1, tb.table_name, tb.column_id - 1
    FROM   user_constraints, USER_CONS_COLUMNS ca, USER_CONS_COLUMNS cb,
           user_tab_cols ta, user_tab_cols tb
    WHERE  user_constraints.table_name = %s AND
           ta.table_name = %s AND
           ta.column_name = ca.column_name AND
           ca.table_name = %s AND
           user_constraints.constraint_name = ca.constraint_name AND
           user_constraints.r_constraint_name = cb.constraint_name AND
           cb.table_name = tb.table_name AND
           cb.column_name = tb.column_name AND
           ca.position = cb.position""", [table_name, table_name, table_name])

        relations = {}
        for row in cursor.fetchall():
            relations[row[0]] = (row[2], row[1].lower())
        return relations

#-----------------------------------------------------------------------------------------------------------------------

    def get_nonunique_indexes(self, cursor, table_name):
        table_name = table_name.upper()
        cursor.execute("""
     select column_name
 from user_indexes i,
      user_ind_columns c
 where c.table_name=%s
    and c.table_name = i.table_name
    and c.index_name = i.index_name
    and uniqueness='NONUNIQUE';""", [table_name])

        indexes = []
        for row in cursor.fetchall():
            indexes.append(row[0].lower())
        return indexes

#-----------------------------------------------------------------------------------------------------------------------

    def get_indexes(self, cursor, table_name):
        """
        Returns a dictionary of fieldname -> infodict for the given table,
        where each infodict is in the format:
            {'primary_key': boolean representing whether it's the primary key,
             'unique': boolean representing whether it's a unique index}
        """
        # This query retrieves each index on the given table, including the
        # first associated field name
        # "We were in the nick of time; you were in great peril!"
        sql = """\
SELECT LOWER(all_tab_cols.column_name) AS column_name,
       CASE user_constraints.constraint_type
           WHEN 'P' THEN 1 ELSE 0
       END AS is_primary_key,
       CASE user_indexes.uniqueness
           WHEN 'UNIQUE' THEN 1 ELSE 0
       END AS is_unique
FROM   all_tab_cols, user_cons_columns, user_constraints, user_ind_columns, user_indexes
WHERE  all_tab_cols.column_name = user_cons_columns.column_name (+)
  AND  all_tab_cols.table_name = user_cons_columns.table_name (+)
  AND  user_cons_columns.constraint_name = user_constraints.constraint_name (+)
  AND  user_constraints.constraint_type (+) = 'P'
  AND  user_ind_columns.column_name (+) = all_tab_cols.column_name
  AND  user_ind_columns.table_name (+) = all_tab_cols.table_name
  AND  user_indexes.uniqueness (+) = 'UNIQUE'
  AND  user_indexes.index_name (+) = user_ind_columns.index_name
  AND  all_tab_cols.table_name = UPPER(%s)
"""
        cursor.execute(sql, [table_name])
        indexes = {}
        for row in cursor.fetchall():
            if not row[0] in indexes:
                indexes[row[0]] = {'primary_key': row[1], 'unique': row[2]}
            else:
                primary = indexes[row[0]]['primary_key']
                unique = indexes[row[0]]['unique']

                if not primary and row[1]:
                    indexes[row[0]]['primary_key'] = row[1]
                if not unique and row[2]:
                    indexes[row[0]]['unique'] = row[2]
        return indexes

#-----------------------------------------------------------------------------------------------------------------------